// Solutions to the Advent Of Code 2024.
// Copyright (C) 2024  Stefan Mller
//
// This program is free software: you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along with
// this program.  If not, see <http://www.gnu.org/licenses/>.

#include <aoc/HoofIt.hpp>

#include <memory>
#include <queue>
#include <vector>

#include <aoc/Point2.hpp>

const std::string HoofIt::getPuzzleName() const
{
    return "Day 10: Hoof It";
}

const std::string HoofIt::getInputFileName() const
{
    return "hoof_it.txt";
}

void HoofIt::finish()
{
    std::queue<Point2> queue{};
    std::vector<std::vector<std::vector<Point2>>> trailEndsMap{};
    trailEndsMap.reserve(lines.size());
    std::vector<std::vector<int>> trailScoreMap{};
    trailScoreMap.reserve(lines.size());

    for (int j = 0; j < lines.size(); j++)
    {
        std::vector<std::vector<Point2>> endsRow{};
        endsRow.reserve(lines[j].size());
        trailEndsMap.push_back(endsRow);
        std::vector<int> scoreRow(lines[j].size(), 0);
        trailScoreMap.push_back(scoreRow);

        for (int i = 0; i < lines[j].size(); i++)
        {
            std::vector<Point2> ends{};
            trailEndsMap[j].push_back(ends);
            if (lines[j][i] == getTrailTopChar())
            {
                Point2 trailEnd{ i, j };
                trailEndsMap[j][i].push_back(trailEnd);
                trailScoreMap[j][i] = 1;
                queue.push(trailEnd);
            }
        }
    }

    while (!queue.empty())
    {
        auto trail = queue.front();
        queue.pop();
        if (getPosition(trail) != getTrailheadChar())
        {
            for (auto direction : Point2::cardinalDirections)
            {
                Point2 p{ trail + direction };
                if (isInBounds(p) && getPosition(p) + 1 == getPosition(trail))
                {
                    if (trailEndsMap[p.y][p.x].empty())
                    {
                        queue.push(p);
                    }
                    addUnique(trailEndsMap[trail.y][trail.x], trailEndsMap[p.y][p.x]);
                    trailScoreMap[p.y][p.x] += trailScoreMap[trail.y][trail.x];
                }
            }
        }
        else
        {
            part1 += trailEndsMap[trail.y][trail.x].size();
            part2 += trailScoreMap[trail.y][trail.x];
        }
    }
}

const char HoofIt::getTrailheadChar() const
{
    return '0';
}

const char HoofIt::getTrailTopChar() const
{
    return '9';
}

void HoofIt::addUnique(const std::vector<Point2>& source, std::vector<Point2>& destination)
{
    for (auto end : source)
    {
        auto isUnique{ true };
        size_t i{ 0 };
        while (isUnique && i < destination.size())
        {
            if (end == destination[i])
            {
                isUnique = false;
            }
            i++;
        }
        if (isUnique)
        {
            destination.push_back(end);
        }
    }
}
